/*
 * Choeps Next Generation
 * 
 * Mark Spencer <markster@marko.net>
 *
 * Copyright(C) 1999, Adtran, Inc.
 * 
 * Distributed under the terms of the GNU General Public License (GPL) Version
 *
 * Agent Handelers for the GUI
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <malloc.h>
#include <stdlib.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <adns.h>
#include "cheops.h"
#include "gui-io.h"
#include "logger.h"
#include "event.h"
#include "ip_utils.h"
#include "cheops-gui.h"
#include "gui-viewspace.h"
#include "cheops-osscan.h"
#include "gui-canvas.h"
#include "gui-utils.h"
#include "gui-settings.h"
#include "gui-dns.h"

#ifdef DEBUG_GUI_HANDLERS
#define DEBUG DEBUG_M
#else
#define DEBUG(a) 
#endif

typedef struct _ip_cache {
	struct _ip_cache *next;
	u32 ip;
	char *name;
} ip_cache;

ip_cache *IP_cache = NULL;

static struct handler {
	int id;
	cheops_event_handler cb;
	char *desc;
} gui_handlers[];

typedef struct _discover_dns {
	char *start;
	char *end;
	u32 start_ip;
	u32 end_ip;
	int flags;
	net_page *np;
	agent *a;
	ip_cache *cache;
	int timeout_id;
} discover_dns;

#define DISCOVER_DNS_FLAGS_NETWORK  1
#define DISCOVER_DNS_FLAGS_USED_DNS 2


void do_host_discover_after_dns(char *ip,char *hostname, void *data);
void do_network_discover_after_dns(char *ip, char *hostname, void *data);


ip_cache *add_ip_cache(u32 ip, char *name)
{
	ip_cache *p = malloc(sizeof(ip_cache));
	
	if(p)
	{
		p->ip = ip;
		p->name = makestring(name);
		p->next = IP_cache;
		IP_cache = p;
	}
	return(p);
}	

char *get_ip_cache(u32 ip)
{
	ip_cache *p, *prev = NULL;
	char *name;
	
	for(p = IP_cache; p; p = p->next)
		if(ip == p->ip)
			break;
		else
			prev = p;
	
	if(p)
	{
		name = p->name;
		
		if(prev)
			prev->next = p->next;
		else
			IP_cache = p->next;
			
		free(p);
		return(name);
	}
	return(NULL);
}
	

void dns_timeout(char *ip, char *hostname, void *data)
{
	discover_dns *dd = data;
	char buf[256];
			
	sprintf(buf,"I could not get an IP address for '%s'. Do you want me to try it again?",dd->start);
	if(make_gnome_yes_no_dialog("DNS lookup error",buf,TRUE, NULL) == 0 )
	{
		//retry the dns
		if(dd->flags & DISCOVER_DNS_FLAGS_NETWORK)
			dns_name_lookup( dd->start, do_network_discover_after_dns, dns_timeout, (void *)dd);
		else
			dns_name_lookup( dd->start, do_host_discover_after_dns, dns_timeout,  (void *)dd);
	}
	else
	{
		if(dd->start)
			free(dd->start);
		if(dd->end)
			free(dd->end);
		free(dd);
	}
}

static int handle_error(event_hdr *h, event *e, agent *a)
{
	clog(LOG_ERROR,"\nAgent reports: Error[%d] - %s", ntohs(e->error_r.error), ((char *)e) + sizeof(error_r));
	return(0);
}

int send_discover_request(discover_dns *dd)
{
	int done = -1;
	
	eh->len = htonl(sizeof(event_hdr) + sizeof(discover_ipv4_e));
	eh->hlen = htons(sizeof(event_hdr));
	eh->type = htons(EVENT_DISCOVER_IPV4);
	eh->flags = dd->flags;
	ee->discover_ipv4_e.start = htonl(dd->start_ip);
	ee->discover_ipv4_e.end = htonl(dd->end_ip);
	ee->discover_ipv4_e.is_name = FALSE;
	ee->discover_ipv4_e.np = dd->np;

	// cache the name so that we can retrieve it when we get a reply
	if(dd->start)
		dd->cache = add_ip_cache(htonl(dd->start_ip), dd->start);
	
	if (event_send(dd->a, eh) < 0) 
		clog(LOG_WARNING, "Unable to send discover event\n");
	else
		done = 0;

	return(done);	
}

void do_host_discover_after_dns(char *ip, char *hostname, void *data)
{
	discover_dns *dd = data;
	struct in_addr addr;
	
	if(inet_aton(ip, &addr))
	{
		DEBUG( printf("Discovering %s (%s)\n", host, ip) );
		dd->start_ip = htonl(addr.s_addr);
		dd->end_ip = dd->start_ip;
		
		send_discover_request(dd);
	}
	if(dd->start)
		free(dd->start);
	if(dd->end)
		free(dd->end);
	
	free(dd);
}

void do_network_discover_after_dns(char *ip, char *hostname, void *data)
{
	int res;
	u32 mask;
	discover_dns *dd = data;
	struct in_addr addr;
		
	if(inet_aton(ip,&addr))
	{
		/* Check the netmask, first to see if it's of the form 255.255.255.0 */
		res = inet_aton(dd->end, (struct in_addr *)&mask);
		mask = ntohl(mask);
		if (!res || (mask < 0x80000000) || !allones(mask))
		{
			if ((atoi(dd->end) > 32) || (atoi(dd->end) < 1))
			{
				/* We don't allow a zero netmask, that would
				   be the entire internet! */
				printf("Invalid netmask: %s\n", dd->end);
				return;
			}
			mask = 0xffffffff;
			mask = mask >> (32-atoi(dd->end));
			mask = mask << (32-atoi(dd->end));
		}
		dd->start_ip = htonl(addr.s_addr);
		dd->start_ip &= mask;
		dd->end_ip = dd->start_ip | ~mask;
		DEBUG( printf("Discovering from %s to ", ip2str(htonl(dd->start_ip))) );
		DEBUG( printf("%s (mask = ", ip2str(htonl(dd->end_ip)) ));
		DEBUG( printf("%s)...\n", ip2str(htonl(mask)) ));
		
		send_discover_request(dd);
	}
	if(dd->start)
		free(dd->start);
	if(dd->end)
		free(dd->end);
	
	free(dd);
}

int do_discover(agent *a, char *host, void *np, int flags)
{
	char *c;
	discover_dns *dd = malloc(sizeof(discover_dns));
	
	dd->start_ip = 0;
	dd->end_ip = 0;
	dd->start = NULL;
	dd->end = NULL;
	dd->flags = 0;
	dd->a = a;
	dd->np = np;
	
	if ((c = strchr(host, '/')))
	{
		/* Network/netmask */
		*c='\0';
		c++;
		
		dd->end = makestring(c);
		dd->flags |= DISCOVER_DNS_FLAGS_NETWORK;
		
		if( 0 == inet_aton(host, (struct in_addr *)&(dd->start_ip)) )
		{
			//is a hostname
			//this is a network address ie: 255.255.255.0
			dd->flags |= DISCOVER_DNS_FLAGS_USED_DNS;

			dd->start = makestring(host);
			dns_name_lookup( dd->start, do_network_discover_after_dns,dns_timeout, (void *)dd);
		}
		else
		{
			//is a ip address
			//this is a network address ie: 255.255.255.0
			dd->start = makestring(host);	
			do_network_discover_after_dns( host ,NULL, dd);
		}
	}
	else
	{
		/* Host */
		if( 0 == inet_aton(host, (struct in_addr *)&dd->start_ip) )
		{
			//hostname
			dd->flags |= DISCOVER_DNS_FLAGS_USED_DNS;
			dd->start = makestring(host);

			dns_name_lookup( dd->start, do_host_discover_after_dns,dns_timeout, (void *)dd);
		}
		else
		{
			//ip address
			do_host_discover_after_dns(host,NULL, dd);
		}
	}
	return(0);
}

void reverse_lookup_cb(char *ip, char *hostname, void *data)
{
	page_object *po = data;
	
	if( 0 == strcmp(inet_ntoa(*(struct in_addr *)&po->ip), ip) )
	{
		
		if(po->name)
			free(po->name);
		
		po->name = makestring(hostname);	
	
		page_object_display(po);
	}
}



int handle_discover_reply(event_hdr *h, event *e, agent *a)
{
	page_object *po;
	char *name = get_ip_cache(e->discover_ipv4_r.ipaddr);
	
	po = add_host_entry_to_net_page(e->discover_ipv4_r.np, e->discover_ipv4_r.ipaddr , name );

	if(options_operating_system_check)
		do_os_scan(a, e->discover_ipv4_r.ipaddr, e->discover_ipv4_r.np, 0);

	if(!name && options_reverse_dns)
		dns_reverse_lookup( inet_ntoa(*(struct in_addr *)&e->discover_ipv4_r.ipaddr), reverse_lookup_cb, NULL, po);

	if(name)
		free(name);
	
	return(1);
}

/******************************************************************************/
/*                              OS detection                                  */
/******************************************************************************/

int do_os_scan(agent *a, int ip, void *np, int flags)
{
	eh->len = htonl(sizeof(event_hdr) + sizeof(os_scan_e));
	eh->hlen = htons(sizeof(event_hdr));
	eh->type = htons(EVENT_OS_SCAN);
	eh->flags = flags;
	ee->os_scan_e.ip = ip;
	ee->os_scan_e.np = np;
	
	if (event_send(a, eh) < 0) 
		clog(LOG_WARNING, "Unable to send OS-scan event\n");
	return(0);
}


int handle_os_scan_reply(event_hdr *h, event *e, agent *a)
{
	char *ptr = (void *)e + sizeof(os_scan_r);
	os_scan_option *opt;
	os_scan_option_os *os;
	os_scan_option_port *port;
	os_stats	*os_data;
	os_port_entry	*os_port;
	int i,j = ntohl(e->os_scan_r.num_options);
	page_object *po;
	
	os_data = malloc(sizeof(os_stats));
	
	for(i = 0; i < j; i++)
	{
		opt = (void *)ptr;
		switch(ntohl(opt->type))
		{
			case OS_SCAN_OPTION_OS:
				os = (void *)opt;
				os_data->os = makestring((char *)&os->string);
				os_data->ports = NULL;
				
				ptr += ntohl(opt->length);
				break;
			
			case OS_SCAN_OPTION_PORT:
				port = (void *)opt;
				os_port = malloc(sizeof(os_port_entry));
				os_port->port = ntohs(port->port_number);
				os_port->protocol = port->protocol;
				os_port->name = NULL;

				add_service(os_port->port, os_port->protocol, port->name);

				os_port->next = os_data->ports;
				os_data->ports = os_port;
				
				ptr += ntohl(opt->length);
				break;
					
			default:
				clog(LOG_ERROR,"\nWhat the h@#$ is this? handle_os_scan_reply\n");
				ptr += ntohl(opt->length);
				break;
		}
	}

	po = page_object_get_by_ip_and_net_page( ntohl(e->os_scan_r.ip), e->os_scan_r.np );
	if(po)
	{
		po->os_data = os_data;
		po->flags |= PAGE_OBJECT_CHANGE_POPUP_LIST;
	
		// redisplay the object
		page_object_display(po);
	}
	return(0);
}


/******************************************************************************/
/*                              DNS Request                                   */
/******************************************************************************/

int do_dns_request_scan(agent *a, char *host, int flags)
{
	eh->len = htonl(sizeof(event_hdr) + sizeof(discover_ipv4_e));
	eh->hlen = htons(sizeof(event_hdr));
	eh->type = htons(EVENT_DNS_QUERY);
	eh->flags = flags;
	strcpy(ee->dns_query_e.name, host);
	
	if (event_send(a, eh) < 0) 
		clog(LOG_WARNING, "Unable to send DNS Query event\n");
	return(0);
}


int handle_dns_query_reply(event_hdr *h, event *e, agent *a)
{

	clog(LOG_NOTICE, "i got a dns reply");		

	return(0);
}





/* *****************************       HANDLERS        ************************* */

struct handler gui_handlers[] = {
	{ REPLY_ERROR,		handle_error,		"Error Reply" },
	{ REPLY_DISCOVER_IPV4,	handle_discover_reply,	"Discover Reply Handler" },
	{ REPLY_OS_SCAN,	handle_os_scan_reply,	"OS Scan Reply Handler" },
	{ REPLY_DNS_QUERY,	handle_dns_query_reply,	"DNS Query Reply Handler" },
	{ -1 },
};

void register_gui_handlers()
{
	struct handler *h = gui_handlers;
	while(h->id > -1) {
		/* Try to register each of our shell items, but don't replace an existing
		   handler */
		if (event_register_handler(h->id, h->cb, 0)) 
			clog(LOG_WARNING, "Unable to register handler for '%s'\n", h->desc);
		h++;
	}
}
